library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ray_unit_tester is
end ray_unit_tester;

architecture ray_test of ray_unit_tester is
	signal clk : std_logic := '0';
	signal col_base : signed (8 downto 0) := (others => '0'); -- should be one of these: -160, -150, ... , -10, 0, 10, ... , 150
	signal row : signed (8 downto 0) := (others => '0'); -- should start at 189 and go down to -50
	signal col_offset : unsigned (3 downto 0) := (others => '0'); -- should start at 0 and go to 9
	signal cam_co_ord_x, cam_co_ord_y, cam_co_ord_z : unsigned (8 downto 0) := (others => '0'); -- coordinates of the camera
	
	-- control signals
	signal new_pixel : std_logic := '0'; -- assert for 1 cycle when starting a new pixel
	signal new_rectangle : std_logic := '0'; -- assert for 1 cycle with each new rectangle
	signal done : std_logic := '0'; -- active high. Wait for this signal after you assert new_rectangle
	signal done_rectangles : std_logic := '0'; -- assert for 1 cycle when you finish all the rectangles
	
	-- rectangles are 7 bytes. x,y,z for corners 1 and 2, and 1 byte for color.
	signal rec_x0, rec_y0, rec_z0, rec_x1, rec_y1, rec_z1 : unsigned (8 downto 0) := (others => '0');
	signal color : unsigned (5 downto 0) := (others => '0');
	
	-- color output
	-- this is the final result.
	signal red, green, blue : unsigned (5 downto 0) := (others => '0');
begin
	unit: entity work.ray_unit port map (
		clk => clk,
		col_base => col_base,
		row => row,
		col_offset => col_offset,
		cam_co_ord_x => cam_co_ord_x,
		cam_co_ord_y => cam_co_ord_y,
		cam_co_ord_z => cam_co_ord_z,
		new_pixel => new_pixel,
		new_rectangle => new_rectangle,
		done => done,
		done_rectangles => done_rectangles,
		rec_x0 => rec_x0,
		rec_y0 => rec_y0,
		rec_z0 => rec_z0,
		rec_x1 => rec_x1,
		rec_y1 => rec_y1,
		rec_z1 => rec_z1,
		color => color,
		red => red,
		green => green,
		blue => blue
	);
	
	process
	begin
		loop
			wait for 10 ns;
			clk <= '1';
			wait for 10 ns;
			clk <= '0';
		end loop;
	end process;
	
	process
	begin
		cam_co_ord_x <= "000000000"; cam_co_ord_y <= "010000000"; cam_co_ord_z <= "000011001";
		col_base <= "101100000"; row <= "111011110"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-159,-34) or (319,224)" severity error;
		assert green = "011111" report "green 011111 (-159,-34) or (319,224)" severity error;
		assert blue = "000101" report "blue 000101 (-159,-34) or (319,224)" severity error;

		col_base <= "101100000"; row <= "111101111"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-158,-17) or (318,207)" severity error;
		assert green = "000000" report "green 000000 (-158,-17) or (318,207)" severity error;
		assert blue = "000110" report "blue 000110 (-158,-17) or (318,207)" severity error;

		col_base <= "101100000"; row <= "000000000"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-157,0) or (317,190)" severity error;
		assert green = "000000" report "green 000000 (-157,0) or (317,190)" severity error;
		assert blue = "001001" report "blue 001001 (-157,0) or (317,190)" severity error;

		col_base <= "101100000"; row <= "000010001"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-156,17) or (316,173)" severity error;
		assert green = "000000" report "green 000000 (-156,17) or (316,173)" severity error;
		assert blue = "001100" report "blue 001100 (-156,17) or (316,173)" severity error;

		col_base <= "101100000"; row <= "000100010"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-155,34) or (315,156)" severity error;
		assert green = "000000" report "green 000000 (-155,34) or (315,156)" severity error;
		assert blue = "001111" report "blue 001111 (-155,34) or (315,156)" severity error;

		col_base <= "101100000"; row <= "000110011"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-154,51) or (314,139)" severity error;
		assert green = "000000" report "green 000000 (-154,51) or (314,139)" severity error;
		assert blue = "010010" report "blue 010010 (-154,51) or (314,139)" severity error;

		col_base <= "101100000"; row <= "001000100"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-153,68) or (313,122)" severity error;
		assert green = "000000" report "green 000000 (-153,68) or (313,122)" severity error;
		assert blue = "010110" report "blue 010110 (-153,68) or (313,122)" severity error;

		col_base <= "101100000"; row <= "001010101"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-152,85) or (312,105)" severity error;
		assert green = "000000" report "green 000000 (-152,85) or (312,105)" severity error;
		assert blue = "011001" report "blue 011001 (-152,85) or (312,105)" severity error;

		col_base <= "101100000"; row <= "001100110"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-151,102) or (311,88)" severity error;
		assert green = "000000" report "green 000000 (-151,102) or (311,88)" severity error;
		assert blue = "011100" report "blue 011100 (-151,102) or (311,88)" severity error;

		col_base <= "101101010"; row <= "001110111"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-150,119) or (310,71)" severity error;
		assert green = "000000" report "green 000000 (-150,119) or (310,71)" severity error;
		assert blue = "011111" report "blue 011111 (-150,119) or (310,71)" severity error;

		col_base <= "101101010"; row <= "010001000"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-149,136) or (309,54)" severity error;
		assert green = "000000" report "green 000000 (-149,136) or (309,54)" severity error;
		assert blue = "100010" report "blue 100010 (-149,136) or (309,54)" severity error;

		col_base <= "101101010"; row <= "010011001"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-148,153) or (308,37)" severity error;
		assert green = "000000" report "green 000000 (-148,153) or (308,37)" severity error;
		assert blue = "100101" report "blue 100101 (-148,153) or (308,37)" severity error;

		col_base <= "101101010"; row <= "010101010"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-147,170) or (307,20)" severity error;
		assert green = "000000" report "green 000000 (-147,170) or (307,20)" severity error;
		assert blue = "101001" report "blue 101001 (-147,170) or (307,20)" severity error;

		col_base <= "101101010"; row <= "010111011"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-146,187) or (306,3)" severity error;
		assert green = "000000" report "green 000000 (-146,187) or (306,3)" severity error;
		assert blue = "101100" report "blue 101100 (-146,187) or (306,3)" severity error;

		col_base <= "101101010"; row <= "111011100"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-144,-36) or (304,226)" severity error;
		assert green = "011111" report "green 011111 (-144,-36) or (304,226)" severity error;
		assert blue = "000101" report "blue 000101 (-144,-36) or (304,226)" severity error;

		col_base <= "101101010"; row <= "111101101"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-143,-19) or (303,209)" severity error;
		assert green = "000000" report "green 000000 (-143,-19) or (303,209)" severity error;
		assert blue = "000101" report "blue 000101 (-143,-19) or (303,209)" severity error;

		col_base <= "101101010"; row <= "111111110"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-142,-2) or (302,192)" severity error;
		assert green = "000000" report "green 000000 (-142,-2) or (302,192)" severity error;
		assert blue = "001001" report "blue 001001 (-142,-2) or (302,192)" severity error;

		col_base <= "101101010"; row <= "000001111"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-141,15) or (301,175)" severity error;
		assert green = "000000" report "green 000000 (-141,15) or (301,175)" severity error;
		assert blue = "001100" report "blue 001100 (-141,15) or (301,175)" severity error;

		col_base <= "101110100"; row <= "000100000"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-140,32) or (300,158)" severity error;
		assert green = "000000" report "green 000000 (-140,32) or (300,158)" severity error;
		assert blue = "001111" report "blue 001111 (-140,32) or (300,158)" severity error;

		col_base <= "101110100"; row <= "000110001"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-139,49) or (299,141)" severity error;
		assert green = "000000" report "green 000000 (-139,49) or (299,141)" severity error;
		assert blue = "010010" report "blue 010010 (-139,49) or (299,141)" severity error;

		col_base <= "101110100"; row <= "001000010"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-138,66) or (298,124)" severity error;
		assert green = "000000" report "green 000000 (-138,66) or (298,124)" severity error;
		assert blue = "010101" report "blue 010101 (-138,66) or (298,124)" severity error;

		col_base <= "101110100"; row <= "001010011"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-137,83) or (297,107)" severity error;
		assert green = "000000" report "green 000000 (-137,83) or (297,107)" severity error;
		assert blue = "011000" report "blue 011000 (-137,83) or (297,107)" severity error;

		col_base <= "101110100"; row <= "001100100"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-136,100) or (296,90)" severity error;
		assert green = "000000" report "green 000000 (-136,100) or (296,90)" severity error;
		assert blue = "011100" report "blue 011100 (-136,100) or (296,90)" severity error;

		col_base <= "101110100"; row <= "001110101"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-135,117) or (295,73)" severity error;
		assert green = "000000" report "green 000000 (-135,117) or (295,73)" severity error;
		assert blue = "011111" report "blue 011111 (-135,117) or (295,73)" severity error;

		col_base <= "101110100"; row <= "010000110"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-134,134) or (294,56)" severity error;
		assert green = "000000" report "green 000000 (-134,134) or (294,56)" severity error;
		assert blue = "100010" report "blue 100010 (-134,134) or (294,56)" severity error;

		col_base <= "101110100"; row <= "010010111"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-133,151) or (293,39)" severity error;
		assert green = "000000" report "green 000000 (-133,151) or (293,39)" severity error;
		assert blue = "100101" report "blue 100101 (-133,151) or (293,39)" severity error;

		col_base <= "101110100"; row <= "010101000"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-132,168) or (292,22)" severity error;
		assert green = "000000" report "green 000000 (-132,168) or (292,22)" severity error;
		assert blue = "101000" report "blue 101000 (-132,168) or (292,22)" severity error;

		col_base <= "101110100"; row <= "010111001"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-131,185) or (291,5)" severity error;
		assert green = "000000" report "green 000000 (-131,185) or (291,5)" severity error;
		assert blue = "101011" report "blue 101011 (-131,185) or (291,5)" severity error;

		col_base <= "101111110"; row <= "111011010"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-129,-38) or (289,228)" severity error;
		assert green = "011111" report "green 011111 (-129,-38) or (289,228)" severity error;
		assert blue = "000101" report "blue 000101 (-129,-38) or (289,228)" severity error;

		col_base <= "101111110"; row <= "111101011"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-128,-21) or (288,211)" severity error;
		assert green = "000000" report "green 000000 (-128,-21) or (288,211)" severity error;
		assert blue = "000001" report "blue 000001 (-128,-21) or (288,211)" severity error;

		col_base <= "101111110"; row <= "111111100"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-127,-4) or (287,194)" severity error;
		assert green = "000000" report "green 000000 (-127,-4) or (287,194)" severity error;
		assert blue = "000010" report "blue 000010 (-127,-4) or (287,194)" severity error;

		col_base <= "101111110"; row <= "000001101"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-126,13) or (286,177)" severity error;
		assert green = "000000" report "green 000000 (-126,13) or (286,177)" severity error;
		assert blue = "000011" report "blue 000011 (-126,13) or (286,177)" severity error;

		col_base <= "101111110"; row <= "000011110"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-125,30) or (285,160)" severity error;
		assert green = "000000" report "green 000000 (-125,30) or (285,160)" severity error;
		assert blue = "000101" report "blue 000101 (-125,30) or (285,160)" severity error;

		col_base <= "101111110"; row <= "000101111"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-124,47) or (284,143)" severity error;
		assert green = "000000" report "green 000000 (-124,47) or (284,143)" severity error;
		assert blue = "000110" report "blue 000110 (-124,47) or (284,143)" severity error;

		col_base <= "101111110"; row <= "001000000"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-123,64) or (283,126)" severity error;
		assert green = "000000" report "green 000000 (-123,64) or (283,126)" severity error;
		assert blue = "000111" report "blue 000111 (-123,64) or (283,126)" severity error;

		col_base <= "101111110"; row <= "001010001"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-122,81) or (282,109)" severity error;
		assert green = "000000" report "green 000000 (-122,81) or (282,109)" severity error;
		assert blue = "001000" report "blue 001000 (-122,81) or (282,109)" severity error;

		col_base <= "101111110"; row <= "001100010"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-121,98) or (281,92)" severity error;
		assert green = "000000" report "green 000000 (-121,98) or (281,92)" severity error;
		assert blue = "001001" report "blue 001001 (-121,98) or (281,92)" severity error;

		col_base <= "110001000"; row <= "001110011"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-120,115) or (280,75)" severity error;
		assert green = "000000" report "green 000000 (-120,115) or (280,75)" severity error;
		assert blue = "001010" report "blue 001010 (-120,115) or (280,75)" severity error;

		col_base <= "110001000"; row <= "010000100"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-119,132) or (279,58)" severity error;
		assert green = "000000" report "green 000000 (-119,132) or (279,58)" severity error;
		assert blue = "001011" report "blue 001011 (-119,132) or (279,58)" severity error;

		col_base <= "110001000"; row <= "010010101"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-118,149) or (278,41)" severity error;
		assert green = "000000" report "green 000000 (-118,149) or (278,41)" severity error;
		assert blue = "001100" report "blue 001100 (-118,149) or (278,41)" severity error;

		col_base <= "110001000"; row <= "010100110"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-117,166) or (277,24)" severity error;
		assert green = "000000" report "green 000000 (-117,166) or (277,24)" severity error;
		assert blue = "001101" report "blue 001101 (-117,166) or (277,24)" severity error;

		col_base <= "110001000"; row <= "010110111"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-116,183) or (276,7)" severity error;
		assert green = "000000" report "green 000000 (-116,183) or (276,7)" severity error;
		assert blue = "001110" report "blue 001110 (-116,183) or (276,7)" severity error;

		col_base <= "110001000"; row <= "111011000"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-114,-40) or (274,230)" severity error;
		assert green = "011111" report "green 011111 (-114,-40) or (274,230)" severity error;
		assert blue = "000000" report "blue 000000 (-114,-40) or (274,230)" severity error;

		col_base <= "110001000"; row <= "111101001"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-113,-23) or (273,213)" severity error;
		assert green = "001111" report "green 001111 (-113,-23) or (273,213)" severity error;
		assert blue = "000000" report "blue 000000 (-113,-23) or (273,213)" severity error;

		col_base <= "110001000"; row <= "111111010"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-112,-6) or (272,196)" severity error;
		assert green = "000000" report "green 000000 (-112,-6) or (272,196)" severity error;
		assert blue = "000010" report "blue 000010 (-112,-6) or (272,196)" severity error;

		col_base <= "110001000"; row <= "000001011"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-111,11) or (271,179)" severity error;
		assert green = "000000" report "green 000000 (-111,11) or (271,179)" severity error;
		assert blue = "000011" report "blue 000011 (-111,11) or (271,179)" severity error;

		col_base <= "110010010"; row <= "000011100"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-110,28) or (270,162)" severity error;
		assert green = "000000" report "green 000000 (-110,28) or (270,162)" severity error;
		assert blue = "000100" report "blue 000100 (-110,28) or (270,162)" severity error;

		col_base <= "110010010"; row <= "000101101"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-109,45) or (269,145)" severity error;
		assert green = "000000" report "green 000000 (-109,45) or (269,145)" severity error;
		assert blue = "000101" report "blue 000101 (-109,45) or (269,145)" severity error;

		col_base <= "110010010"; row <= "000111110"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-108,62) or (268,128)" severity error;
		assert green = "000000" report "green 000000 (-108,62) or (268,128)" severity error;
		assert blue = "000111" report "blue 000111 (-108,62) or (268,128)" severity error;

		col_base <= "110010010"; row <= "001001111"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-107,79) or (267,111)" severity error;
		assert green = "000000" report "green 000000 (-107,79) or (267,111)" severity error;
		assert blue = "001000" report "blue 001000 (-107,79) or (267,111)" severity error;

		col_base <= "110010010"; row <= "001100000"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-106,96) or (266,94)" severity error;
		assert green = "000000" report "green 000000 (-106,96) or (266,94)" severity error;
		assert blue = "001001" report "blue 001001 (-106,96) or (266,94)" severity error;

		col_base <= "110010010"; row <= "001110001"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-105,113) or (265,77)" severity error;
		assert green = "000000" report "green 000000 (-105,113) or (265,77)" severity error;
		assert blue = "001010" report "blue 001010 (-105,113) or (265,77)" severity error;

		col_base <= "110010010"; row <= "010000010"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-104,130) or (264,60)" severity error;
		assert green = "000000" report "green 000000 (-104,130) or (264,60)" severity error;
		assert blue = "001011" report "blue 001011 (-104,130) or (264,60)" severity error;

		col_base <= "110010010"; row <= "010010011"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-103,147) or (263,43)" severity error;
		assert green = "000000" report "green 000000 (-103,147) or (263,43)" severity error;
		assert blue = "001100" report "blue 001100 (-103,147) or (263,43)" severity error;

		col_base <= "110010010"; row <= "010100100"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-102,164) or (262,26)" severity error;
		assert green = "000000" report "green 000000 (-102,164) or (262,26)" severity error;
		assert blue = "001101" report "blue 001101 (-102,164) or (262,26)" severity error;

		col_base <= "110010010"; row <= "010110101"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-101,181) or (261,9)" severity error;
		assert green = "000000" report "green 000000 (-101,181) or (261,9)" severity error;
		assert blue = "001110" report "blue 001110 (-101,181) or (261,9)" severity error;

		col_base <= "110011100"; row <= "111010110"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-99,-42) or (259,232)" severity error;
		assert green = "011111" report "green 011111 (-99,-42) or (259,232)" severity error;
		assert blue = "000000" report "blue 000000 (-99,-42) or (259,232)" severity error;

		col_base <= "110011100"; row <= "111100111"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-98,-25) or (258,215)" severity error;
		assert green = "011111" report "green 011111 (-98,-25) or (258,215)" severity error;
		assert blue = "000100" report "blue 000100 (-98,-25) or (258,215)" severity error;

		col_base <= "110011100"; row <= "111111000"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-97,-8) or (257,198)" severity error;
		assert green = "000000" report "green 000000 (-97,-8) or (257,198)" severity error;
		assert blue = "000010" report "blue 000010 (-97,-8) or (257,198)" severity error;

		col_base <= "110011100"; row <= "000001001"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-96,9) or (256,181)" severity error;
		assert green = "000000" report "green 000000 (-96,9) or (256,181)" severity error;
		assert blue = "000011" report "blue 000011 (-96,9) or (256,181)" severity error;

		col_base <= "110011100"; row <= "000011010"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-95,26) or (255,164)" severity error;
		assert green = "000000" report "green 000000 (-95,26) or (255,164)" severity error;
		assert blue = "000100" report "blue 000100 (-95,26) or (255,164)" severity error;

		col_base <= "110011100"; row <= "000101011"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-94,43) or (254,147)" severity error;
		assert green = "000000" report "green 000000 (-94,43) or (254,147)" severity error;
		assert blue = "000101" report "blue 000101 (-94,43) or (254,147)" severity error;

		col_base <= "110011100"; row <= "000111100"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-93,60) or (253,130)" severity error;
		assert green = "000000" report "green 000000 (-93,60) or (253,130)" severity error;
		assert blue = "000110" report "blue 000110 (-93,60) or (253,130)" severity error;

		col_base <= "110011100"; row <= "001001101"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-92,77) or (252,113)" severity error;
		assert green = "000000" report "green 000000 (-92,77) or (252,113)" severity error;
		assert blue = "000111" report "blue 000111 (-92,77) or (252,113)" severity error;

		col_base <= "110011100"; row <= "001011110"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-91,94) or (251,96)" severity error;
		assert green = "000000" report "green 000000 (-91,94) or (251,96)" severity error;
		assert blue = "001001" report "blue 001001 (-91,94) or (251,96)" severity error;

		col_base <= "110100110"; row <= "001101111"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-90,111) or (250,79)" severity error;
		assert green = "000000" report "green 000000 (-90,111) or (250,79)" severity error;
		assert blue = "001010" report "blue 001010 (-90,111) or (250,79)" severity error;

		col_base <= "110100110"; row <= "010000000"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-89,128) or (249,62)" severity error;
		assert green = "000000" report "green 000000 (-89,128) or (249,62)" severity error;
		assert blue = "001011" report "blue 001011 (-89,128) or (249,62)" severity error;

		col_base <= "110100110"; row <= "010010001"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-88,145) or (248,45)" severity error;
		assert green = "000000" report "green 000000 (-88,145) or (248,45)" severity error;
		assert blue = "001100" report "blue 001100 (-88,145) or (248,45)" severity error;

		col_base <= "110100110"; row <= "010100010"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-87,162) or (247,28)" severity error;
		assert green = "000000" report "green 000000 (-87,162) or (247,28)" severity error;
		assert blue = "001101" report "blue 001101 (-87,162) or (247,28)" severity error;

		col_base <= "110100110"; row <= "010110011"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-86,179) or (246,11)" severity error;
		assert green = "000000" report "green 000000 (-86,179) or (246,11)" severity error;
		assert blue = "001110" report "blue 001110 (-86,179) or (246,11)" severity error;

		col_base <= "110100110"; row <= "111010100"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-84,-44) or (244,234)" severity error;
		assert green = "011111" report "green 011111 (-84,-44) or (244,234)" severity error;
		assert blue = "000000" report "blue 000000 (-84,-44) or (244,234)" severity error;

		col_base <= "110100110"; row <= "111100101"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-83,-27) or (243,217)" severity error;
		assert green = "011111" report "green 011111 (-83,-27) or (243,217)" severity error;
		assert blue = "000000" report "blue 000000 (-83,-27) or (243,217)" severity error;

		col_base <= "110100110"; row <= "111110110"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-82,-10) or (242,200)" severity error;
		assert green = "000000" report "green 000000 (-82,-10) or (242,200)" severity error;
		assert blue = "000010" report "blue 000010 (-82,-10) or (242,200)" severity error;

		col_base <= "110100110"; row <= "000000111"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-81,7) or (241,183)" severity error;
		assert green = "000000" report "green 000000 (-81,7) or (241,183)" severity error;
		assert blue = "000011" report "blue 000011 (-81,7) or (241,183)" severity error;

		col_base <= "110110000"; row <= "000011000"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-80,24) or (240,166)" severity error;
		assert green = "000000" report "green 000000 (-80,24) or (240,166)" severity error;
		assert blue = "000100" report "blue 000100 (-80,24) or (240,166)" severity error;

		col_base <= "110110000"; row <= "000101001"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-79,41) or (239,149)" severity error;
		assert green = "000000" report "green 000000 (-79,41) or (239,149)" severity error;
		assert blue = "000101" report "blue 000101 (-79,41) or (239,149)" severity error;

		col_base <= "110110000"; row <= "000111010"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-78,58) or (238,132)" severity error;
		assert green = "000000" report "green 000000 (-78,58) or (238,132)" severity error;
		assert blue = "000110" report "blue 000110 (-78,58) or (238,132)" severity error;

		col_base <= "110110000"; row <= "001001011"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-77,75) or (237,115)" severity error;
		assert green = "000000" report "green 000000 (-77,75) or (237,115)" severity error;
		assert blue = "000111" report "blue 000111 (-77,75) or (237,115)" severity error;

		col_base <= "110110000"; row <= "001011100"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-76,92) or (236,98)" severity error;
		assert green = "000000" report "green 000000 (-76,92) or (236,98)" severity error;
		assert blue = "001000" report "blue 001000 (-76,92) or (236,98)" severity error;

		col_base <= "110110000"; row <= "001101101"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-75,109) or (235,81)" severity error;
		assert green = "000000" report "green 000000 (-75,109) or (235,81)" severity error;
		assert blue = "001001" report "blue 001001 (-75,109) or (235,81)" severity error;

		col_base <= "110110000"; row <= "001111110"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-74,126) or (234,64)" severity error;
		assert green = "000000" report "green 000000 (-74,126) or (234,64)" severity error;
		assert blue = "001011" report "blue 001011 (-74,126) or (234,64)" severity error;

		col_base <= "110110000"; row <= "010001111"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-73,143) or (233,47)" severity error;
		assert green = "000000" report "green 000000 (-73,143) or (233,47)" severity error;
		assert blue = "001100" report "blue 001100 (-73,143) or (233,47)" severity error;

		col_base <= "110110000"; row <= "010100000"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-72,160) or (232,30)" severity error;
		assert green = "000000" report "green 000000 (-72,160) or (232,30)" severity error;
		assert blue = "001101" report "blue 001101 (-72,160) or (232,30)" severity error;

		col_base <= "110110000"; row <= "010110001"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-71,177) or (231,13)" severity error;
		assert green = "000000" report "green 000000 (-71,177) or (231,13)" severity error;
		assert blue = "001110" report "blue 001110 (-71,177) or (231,13)" severity error;

		col_base <= "110111010"; row <= "111010010"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-69,-46) or (229,236)" severity error;
		assert green = "011111" report "green 011111 (-69,-46) or (229,236)" severity error;
		assert blue = "000000" report "blue 000000 (-69,-46) or (229,236)" severity error;

		col_base <= "110111010"; row <= "111100011"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-68,-29) or (228,219)" severity error;
		assert green = "011111" report "green 011111 (-68,-29) or (228,219)" severity error;
		assert blue = "000000" report "blue 000000 (-68,-29) or (228,219)" severity error;

		col_base <= "110111010"; row <= "111110100"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-67,-12) or (227,202)" severity error;
		assert green = "000000" report "green 000000 (-67,-12) or (227,202)" severity error;
		assert blue = "000010" report "blue 000010 (-67,-12) or (227,202)" severity error;

		col_base <= "110111010"; row <= "000000101"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-66,5) or (226,185)" severity error;
		assert green = "000000" report "green 000000 (-66,5) or (226,185)" severity error;
		assert blue = "000011" report "blue 000011 (-66,5) or (226,185)" severity error;

		col_base <= "110111010"; row <= "000010110"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-65,22) or (225,168)" severity error;
		assert green = "000000" report "green 000000 (-65,22) or (225,168)" severity error;
		assert blue = "000100" report "blue 000100 (-65,22) or (225,168)" severity error;

		col_base <= "110111010"; row <= "000100111"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-64,39) or (224,151)" severity error;
		assert green = "000000" report "green 000000 (-64,39) or (224,151)" severity error;
		assert blue = "000101" report "blue 000101 (-64,39) or (224,151)" severity error;

		col_base <= "110111010"; row <= "000111000"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-63,56) or (223,134)" severity error;
		assert green = "000000" report "green 000000 (-63,56) or (223,134)" severity error;
		assert blue = "000110" report "blue 000110 (-63,56) or (223,134)" severity error;

		col_base <= "110111010"; row <= "001001001"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-62,73) or (222,117)" severity error;
		assert green = "000000" report "green 000000 (-62,73) or (222,117)" severity error;
		assert blue = "000111" report "blue 000111 (-62,73) or (222,117)" severity error;

		col_base <= "110111010"; row <= "001011010"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-61,90) or (221,100)" severity error;
		assert green = "000000" report "green 000000 (-61,90) or (221,100)" severity error;
		assert blue = "001000" report "blue 001000 (-61,90) or (221,100)" severity error;

		col_base <= "111000100"; row <= "001101011"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-60,107) or (220,83)" severity error;
		assert green = "000000" report "green 000000 (-60,107) or (220,83)" severity error;
		assert blue = "001001" report "blue 001001 (-60,107) or (220,83)" severity error;

		col_base <= "111000100"; row <= "001111100"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-59,124) or (219,66)" severity error;
		assert green = "000000" report "green 000000 (-59,124) or (219,66)" severity error;
		assert blue = "001010" report "blue 001010 (-59,124) or (219,66)" severity error;

		col_base <= "111000100"; row <= "010001101"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-58,141) or (218,49)" severity error;
		assert green = "000000" report "green 000000 (-58,141) or (218,49)" severity error;
		assert blue = "001011" report "blue 001011 (-58,141) or (218,49)" severity error;

		col_base <= "111000100"; row <= "010011110"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-57,158) or (217,32)" severity error;
		assert green = "000000" report "green 000000 (-57,158) or (217,32)" severity error;
		assert blue = "001101" report "blue 001101 (-57,158) or (217,32)" severity error;

		col_base <= "111000100"; row <= "010101111"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-56,175) or (216,15)" severity error;
		assert green = "000000" report "green 000000 (-56,175) or (216,15)" severity error;
		assert blue = "001110" report "blue 001110 (-56,175) or (216,15)" severity error;

		col_base <= "111000100"; row <= "111010000"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-54,-48) or (214,238)" severity error;
		assert green = "011111" report "green 011111 (-54,-48) or (214,238)" severity error;
		assert blue = "000000" report "blue 000000 (-54,-48) or (214,238)" severity error;

		col_base <= "111000100"; row <= "111100001"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-53,-31) or (213,221)" severity error;
		assert green = "011111" report "green 011111 (-53,-31) or (213,221)" severity error;
		assert blue = "000000" report "blue 000000 (-53,-31) or (213,221)" severity error;

		col_base <= "111000100"; row <= "111110010"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-52,-14) or (212,204)" severity error;
		assert green = "001111" report "green 001111 (-52,-14) or (212,204)" severity error;
		assert blue = "000000" report "blue 000000 (-52,-14) or (212,204)" severity error;

		col_base <= "111000100"; row <= "000000011"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-51,3) or (211,187)" severity error;
		assert green = "000000" report "green 000000 (-51,3) or (211,187)" severity error;
		assert blue = "000011" report "blue 000011 (-51,3) or (211,187)" severity error;

		col_base <= "111001110"; row <= "000010100"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-50,20) or (210,170)" severity error;
		assert green = "000000" report "green 000000 (-50,20) or (210,170)" severity error;
		assert blue = "000100" report "blue 000100 (-50,20) or (210,170)" severity error;

		col_base <= "111001110"; row <= "000100101"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-49,37) or (209,153)" severity error;
		assert green = "000000" report "green 000000 (-49,37) or (209,153)" severity error;
		assert blue = "000101" report "blue 000101 (-49,37) or (209,153)" severity error;

		col_base <= "111001110"; row <= "000110110"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-48,54) or (208,136)" severity error;
		assert green = "000000" report "green 000000 (-48,54) or (208,136)" severity error;
		assert blue = "000110" report "blue 000110 (-48,54) or (208,136)" severity error;

		col_base <= "111001110"; row <= "001000111"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-47,71) or (207,119)" severity error;
		assert green = "000000" report "green 000000 (-47,71) or (207,119)" severity error;
		assert blue = "000111" report "blue 000111 (-47,71) or (207,119)" severity error;

		col_base <= "111001110"; row <= "001011000"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-46,88) or (206,102)" severity error;
		assert green = "000000" report "green 000000 (-46,88) or (206,102)" severity error;
		assert blue = "001000" report "blue 001000 (-46,88) or (206,102)" severity error;

		col_base <= "111001110"; row <= "001101001"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-45,105) or (205,85)" severity error;
		assert green = "000000" report "green 000000 (-45,105) or (205,85)" severity error;
		assert blue = "001001" report "blue 001001 (-45,105) or (205,85)" severity error;

		col_base <= "111001110"; row <= "001111010"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-44,122) or (204,68)" severity error;
		assert green = "000000" report "green 000000 (-44,122) or (204,68)" severity error;
		assert blue = "001010" report "blue 001010 (-44,122) or (204,68)" severity error;

		col_base <= "111001110"; row <= "010001011"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-43,139) or (203,51)" severity error;
		assert green = "000000" report "green 000000 (-43,139) or (203,51)" severity error;
		assert blue = "001011" report "blue 001011 (-43,139) or (203,51)" severity error;

		col_base <= "111001110"; row <= "010011100"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-42,156) or (202,34)" severity error;
		assert green = "000000" report "green 000000 (-42,156) or (202,34)" severity error;
		assert blue = "001100" report "blue 001100 (-42,156) or (202,34)" severity error;

		col_base <= "111001110"; row <= "010101101"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-41,173) or (201,17)" severity error;
		assert green = "000000" report "green 000000 (-41,173) or (201,17)" severity error;
		assert blue = "001101" report "blue 001101 (-41,173) or (201,17)" severity error;

		col_base <= "111011000"; row <= "111001110"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-39,-50) or (199,240)" severity error;
		assert green = "011111" report "green 011111 (-39,-50) or (199,240)" severity error;
		assert blue = "000000" report "blue 000000 (-39,-50) or (199,240)" severity error;

		col_base <= "111011000"; row <= "111011111"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-38,-33) or (198,223)" severity error;
		assert green = "011111" report "green 011111 (-38,-33) or (198,223)" severity error;
		assert blue = "000000" report "blue 000000 (-38,-33) or (198,223)" severity error;

		col_base <= "111011000"; row <= "111110000"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "101111" report "red 101111 (-37,-16) or (197,206)" severity error;
		assert green = "001111" report "green 001111 (-37,-16) or (197,206)" severity error;
		assert blue = "000000" report "blue 000000 (-37,-16) or (197,206)" severity error;

		col_base <= "111011000"; row <= "000000001"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-36,1) or (196,189)" severity error;
		assert green = "000000" report "green 000000 (-36,1) or (196,189)" severity error;
		assert blue = "000011" report "blue 000011 (-36,1) or (196,189)" severity error;

		col_base <= "111011000"; row <= "000010010"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-35,18) or (195,172)" severity error;
		assert green = "000000" report "green 000000 (-35,18) or (195,172)" severity error;
		assert blue = "000100" report "blue 000100 (-35,18) or (195,172)" severity error;

		col_base <= "111011000"; row <= "000100011"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-34,35) or (194,155)" severity error;
		assert green = "000000" report "green 000000 (-34,35) or (194,155)" severity error;
		assert blue = "000101" report "blue 000101 (-34,35) or (194,155)" severity error;

		col_base <= "111011000"; row <= "000110100"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-33,52) or (193,138)" severity error;
		assert green = "000000" report "green 000000 (-33,52) or (193,138)" severity error;
		assert blue = "000110" report "blue 000110 (-33,52) or (193,138)" severity error;

		col_base <= "111011000"; row <= "001000101"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-32,69) or (192,121)" severity error;
		assert green = "000000" report "green 000000 (-32,69) or (192,121)" severity error;
		assert blue = "000111" report "blue 000111 (-32,69) or (192,121)" severity error;

		col_base <= "111011000"; row <= "001010110"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-31,86) or (191,104)" severity error;
		assert green = "000000" report "green 000000 (-31,86) or (191,104)" severity error;
		assert blue = "001000" report "blue 001000 (-31,86) or (191,104)" severity error;

		col_base <= "111100010"; row <= "001100111"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-30,103) or (190,87)" severity error;
		assert green = "000000" report "green 000000 (-30,103) or (190,87)" severity error;
		assert blue = "001001" report "blue 001001 (-30,103) or (190,87)" severity error;

		col_base <= "111100010"; row <= "001111000"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-29,120) or (189,70)" severity error;
		assert green = "000000" report "green 000000 (-29,120) or (189,70)" severity error;
		assert blue = "001010" report "blue 001010 (-29,120) or (189,70)" severity error;

		col_base <= "111100010"; row <= "010001001"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-28,137) or (188,53)" severity error;
		assert green = "000000" report "green 000000 (-28,137) or (188,53)" severity error;
		assert blue = "001011" report "blue 001011 (-28,137) or (188,53)" severity error;

		col_base <= "111100010"; row <= "010011010"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-27,154) or (187,36)" severity error;
		assert green = "000000" report "green 000000 (-27,154) or (187,36)" severity error;
		assert blue = "100110" report "blue 100110 (-27,154) or (187,36)" severity error;

		col_base <= "111100010"; row <= "010101011"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-26,171) or (186,19)" severity error;
		assert green = "000000" report "green 000000 (-26,171) or (186,19)" severity error;
		assert blue = "101001" report "blue 101001 (-26,171) or (186,19)" severity error;

		col_base <= "111100010"; row <= "010111100"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-25,188) or (185,2)" severity error;
		assert green = "000000" report "green 000000 (-25,188) or (185,2)" severity error;
		assert blue = "101100" report "blue 101100 (-25,188) or (185,2)" severity error;

		col_base <= "111100010"; row <= "111011101"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-23,-35) or (183,225)" severity error;
		assert green = "011111" report "green 011111 (-23,-35) or (183,225)" severity error;
		assert blue = "000101" report "blue 000101 (-23,-35) or (183,225)" severity error;

		col_base <= "111100010"; row <= "111101110"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-22,-18) or (182,208)" severity error;
		assert green = "011111" report "green 011111 (-22,-18) or (182,208)" severity error;
		assert blue = "000100" report "blue 000100 (-22,-18) or (182,208)" severity error;

		col_base <= "111100010"; row <= "111111111"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-21,-1) or (181,191)" severity error;
		assert green = "000000" report "green 000000 (-21,-1) or (181,191)" severity error;
		assert blue = "001001" report "blue 001001 (-21,-1) or (181,191)" severity error;

		col_base <= "111101100"; row <= "000010000"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-20,16) or (180,174)" severity error;
		assert green = "000000" report "green 000000 (-20,16) or (180,174)" severity error;
		assert blue = "001100" report "blue 001100 (-20,16) or (180,174)" severity error;

		col_base <= "111101100"; row <= "000100001"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-19,33) or (179,157)" severity error;
		assert green = "000000" report "green 000000 (-19,33) or (179,157)" severity error;
		assert blue = "001111" report "blue 001111 (-19,33) or (179,157)" severity error;

		col_base <= "111101100"; row <= "000110010"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-18,50) or (178,140)" severity error;
		assert green = "000000" report "green 000000 (-18,50) or (178,140)" severity error;
		assert blue = "010010" report "blue 010010 (-18,50) or (178,140)" severity error;

		col_base <= "111101100"; row <= "001000011"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-17,67) or (177,123)" severity error;
		assert green = "000000" report "green 000000 (-17,67) or (177,123)" severity error;
		assert blue = "010101" report "blue 010101 (-17,67) or (177,123)" severity error;

		col_base <= "111101100"; row <= "001010100"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-16,84) or (176,106)" severity error;
		assert green = "000000" report "green 000000 (-16,84) or (176,106)" severity error;
		assert blue = "011001" report "blue 011001 (-16,84) or (176,106)" severity error;

		col_base <= "111101100"; row <= "001100101"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-15,101) or (175,89)" severity error;
		assert green = "000000" report "green 000000 (-15,101) or (175,89)" severity error;
		assert blue = "011100" report "blue 011100 (-15,101) or (175,89)" severity error;

		col_base <= "111101100"; row <= "001110110"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-14,118) or (174,72)" severity error;
		assert green = "000000" report "green 000000 (-14,118) or (174,72)" severity error;
		assert blue = "011111" report "blue 011111 (-14,118) or (174,72)" severity error;

		col_base <= "111101100"; row <= "010000111"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-13,135) or (173,55)" severity error;
		assert green = "000000" report "green 000000 (-13,135) or (173,55)" severity error;
		assert blue = "100010" report "blue 100010 (-13,135) or (173,55)" severity error;

		col_base <= "111101100"; row <= "010011000"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-12,152) or (172,38)" severity error;
		assert green = "000000" report "green 000000 (-12,152) or (172,38)" severity error;
		assert blue = "100101" report "blue 100101 (-12,152) or (172,38)" severity error;

		col_base <= "111101100"; row <= "010101001"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-11,169) or (171,21)" severity error;
		assert green = "000000" report "green 000000 (-11,169) or (171,21)" severity error;
		assert blue = "101000" report "blue 101000 (-11,169) or (171,21)" severity error;

		col_base <= "111110110"; row <= "010111010"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-10,186) or (170,4)" severity error;
		assert green = "000000" report "green 000000 (-10,186) or (170,4)" severity error;
		assert blue = "101100" report "blue 101100 (-10,186) or (170,4)" severity error;

		col_base <= "111110110"; row <= "111011011"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-8,-37) or (168,227)" severity error;
		assert green = "011111" report "green 011111 (-8,-37) or (168,227)" severity error;
		assert blue = "000101" report "blue 000101 (-8,-37) or (168,227)" severity error;

		col_base <= "111110110"; row <= "111101100"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (-7,-20) or (167,210)" severity error;
		assert green = "011111" report "green 011111 (-7,-20) or (167,210)" severity error;
		assert blue = "000100" report "blue 000100 (-7,-20) or (167,210)" severity error;

		col_base <= "111110110"; row <= "111111101"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-6,-3) or (166,193)" severity error;
		assert green = "000000" report "green 000000 (-6,-3) or (166,193)" severity error;
		assert blue = "001000" report "blue 001000 (-6,-3) or (166,193)" severity error;

		col_base <= "111110110"; row <= "000001110"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-5,14) or (165,176)" severity error;
		assert green = "000000" report "green 000000 (-5,14) or (165,176)" severity error;
		assert blue = "001100" report "blue 001100 (-5,14) or (165,176)" severity error;

		col_base <= "111110110"; row <= "000011111"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-4,31) or (164,159)" severity error;
		assert green = "000000" report "green 000000 (-4,31) or (164,159)" severity error;
		assert blue = "001111" report "blue 001111 (-4,31) or (164,159)" severity error;

		col_base <= "111110110"; row <= "000110000"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-3,48) or (163,142)" severity error;
		assert green = "000000" report "green 000000 (-3,48) or (163,142)" severity error;
		assert blue = "010010" report "blue 010010 (-3,48) or (163,142)" severity error;

		col_base <= "111110110"; row <= "001000001"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-2,65) or (162,125)" severity error;
		assert green = "000000" report "green 000000 (-2,65) or (162,125)" severity error;
		assert blue = "010101" report "blue 010101 (-2,65) or (162,125)" severity error;

		col_base <= "111110110"; row <= "001010010"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (-1,82) or (161,108)" severity error;
		assert green = "000000" report "green 000000 (-1,82) or (161,108)" severity error;
		assert blue = "011000" report "blue 011000 (-1,82) or (161,108)" severity error;

		col_base <= "000000000"; row <= "001100011"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (0,99) or (160,91)" severity error;
		assert green = "000000" report "green 000000 (0,99) or (160,91)" severity error;
		assert blue = "011011" report "blue 011011 (0,99) or (160,91)" severity error;

		col_base <= "000000000"; row <= "001110100"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (1,116) or (159,74)" severity error;
		assert green = "000000" report "green 000000 (1,116) or (159,74)" severity error;
		assert blue = "011111" report "blue 011111 (1,116) or (159,74)" severity error;

		col_base <= "000000000"; row <= "010000101"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (2,133) or (158,57)" severity error;
		assert green = "000000" report "green 000000 (2,133) or (158,57)" severity error;
		assert blue = "100010" report "blue 100010 (2,133) or (158,57)" severity error;

		col_base <= "000000000"; row <= "010010110"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (3,150) or (157,40)" severity error;
		assert green = "000000" report "green 000000 (3,150) or (157,40)" severity error;
		assert blue = "100101" report "blue 100101 (3,150) or (157,40)" severity error;

		col_base <= "000000000"; row <= "010100111"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (4,167) or (156,23)" severity error;
		assert green = "000000" report "green 000000 (4,167) or (156,23)" severity error;
		assert blue = "101000" report "blue 101000 (4,167) or (156,23)" severity error;

		col_base <= "000000000"; row <= "010111000"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (5,184) or (155,6)" severity error;
		assert green = "000000" report "green 000000 (5,184) or (155,6)" severity error;
		assert blue = "101011" report "blue 101011 (5,184) or (155,6)" severity error;

		col_base <= "000000000"; row <= "111011001"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (7,-39) or (153,229)" severity error;
		assert green = "011111" report "green 011111 (7,-39) or (153,229)" severity error;
		assert blue = "000101" report "blue 000101 (7,-39) or (153,229)" severity error;

		col_base <= "000000000"; row <= "111101010"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (8,-22) or (152,212)" severity error;
		assert green = "011111" report "green 011111 (8,-22) or (152,212)" severity error;
		assert blue = "000100" report "blue 000100 (8,-22) or (152,212)" severity error;

		col_base <= "000000000"; row <= "111111011"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (9,-5) or (151,195)" severity error;
		assert green = "000000" report "green 000000 (9,-5) or (151,195)" severity error;
		assert blue = "001000" report "blue 001000 (9,-5) or (151,195)" severity error;

		col_base <= "000001010"; row <= "000001100"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (10,12) or (150,178)" severity error;
		assert green = "000000" report "green 000000 (10,12) or (150,178)" severity error;
		assert blue = "001011" report "blue 001011 (10,12) or (150,178)" severity error;

		col_base <= "000001010"; row <= "000011101"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (11,29) or (149,161)" severity error;
		assert green = "000000" report "green 000000 (11,29) or (149,161)" severity error;
		assert blue = "001110" report "blue 001110 (11,29) or (149,161)" severity error;

		col_base <= "000001010"; row <= "000101110"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (12,46) or (148,144)" severity error;
		assert green = "000000" report "green 000000 (12,46) or (148,144)" severity error;
		assert blue = "010010" report "blue 010010 (12,46) or (148,144)" severity error;

		col_base <= "000001010"; row <= "000111111"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (13,63) or (147,127)" severity error;
		assert green = "000000" report "green 000000 (13,63) or (147,127)" severity error;
		assert blue = "010101" report "blue 010101 (13,63) or (147,127)" severity error;

		col_base <= "000001010"; row <= "001010000"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (14,80) or (146,110)" severity error;
		assert green = "000000" report "green 000000 (14,80) or (146,110)" severity error;
		assert blue = "011000" report "blue 011000 (14,80) or (146,110)" severity error;

		col_base <= "000001010"; row <= "001100001"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (15,97) or (145,93)" severity error;
		assert green = "000000" report "green 000000 (15,97) or (145,93)" severity error;
		assert blue = "011011" report "blue 011011 (15,97) or (145,93)" severity error;

		col_base <= "000001010"; row <= "001110010"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (16,114) or (144,76)" severity error;
		assert green = "000000" report "green 000000 (16,114) or (144,76)" severity error;
		assert blue = "011110" report "blue 011110 (16,114) or (144,76)" severity error;

		col_base <= "000001010"; row <= "010000011"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (17,131) or (143,59)" severity error;
		assert green = "000000" report "green 000000 (17,131) or (143,59)" severity error;
		assert blue = "100001" report "blue 100001 (17,131) or (143,59)" severity error;

		col_base <= "000001010"; row <= "010010100"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (18,148) or (142,42)" severity error;
		assert green = "000000" report "green 000000 (18,148) or (142,42)" severity error;
		assert blue = "100101" report "blue 100101 (18,148) or (142,42)" severity error;

		col_base <= "000001010"; row <= "010100101"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (19,165) or (141,25)" severity error;
		assert green = "000000" report "green 000000 (19,165) or (141,25)" severity error;
		assert blue = "101000" report "blue 101000 (19,165) or (141,25)" severity error;

		col_base <= "000010100"; row <= "010110110"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (20,182) or (140,8)" severity error;
		assert green = "000000" report "green 000000 (20,182) or (140,8)" severity error;
		assert blue = "101011" report "blue 101011 (20,182) or (140,8)" severity error;

		col_base <= "000010100"; row <= "111010111"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (22,-41) or (138,231)" severity error;
		assert green = "011111" report "green 011111 (22,-41) or (138,231)" severity error;
		assert blue = "000101" report "blue 000101 (22,-41) or (138,231)" severity error;

		col_base <= "000010100"; row <= "111101000"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (23,-24) or (137,214)" severity error;
		assert green = "011111" report "green 011111 (23,-24) or (137,214)" severity error;
		assert blue = "000100" report "blue 000100 (23,-24) or (137,214)" severity error;

		col_base <= "000010100"; row <= "111111001"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (24,-7) or (136,197)" severity error;
		assert green = "000000" report "green 000000 (24,-7) or (136,197)" severity error;
		assert blue = "000111" report "blue 000111 (24,-7) or (136,197)" severity error;

		col_base <= "000010100"; row <= "000001010"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (25,10) or (135,180)" severity error;
		assert green = "000000" report "green 000000 (25,10) or (135,180)" severity error;
		assert blue = "001011" report "blue 001011 (25,10) or (135,180)" severity error;

		col_base <= "000010100"; row <= "000011011"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (26,27) or (134,163)" severity error;
		assert green = "000000" report "green 000000 (26,27) or (134,163)" severity error;
		assert blue = "001110" report "blue 001110 (26,27) or (134,163)" severity error;

		col_base <= "000010100"; row <= "000101100"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (27,44) or (133,146)" severity error;
		assert green = "000000" report "green 000000 (27,44) or (133,146)" severity error;
		assert blue = "010001" report "blue 010001 (27,44) or (133,146)" severity error;

		col_base <= "000010100"; row <= "000111101"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (28,61) or (132,129)" severity error;
		assert green = "000000" report "green 000000 (28,61) or (132,129)" severity error;
		assert blue = "010100" report "blue 010100 (28,61) or (132,129)" severity error;

		col_base <= "000010100"; row <= "001001110"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (29,78) or (131,112)" severity error;
		assert green = "000000" report "green 000000 (29,78) or (131,112)" severity error;
		assert blue = "011000" report "blue 011000 (29,78) or (131,112)" severity error;

		col_base <= "000011110"; row <= "001011111"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (30,95) or (130,95)" severity error;
		assert green = "000000" report "green 000000 (30,95) or (130,95)" severity error;
		assert blue = "011011" report "blue 011011 (30,95) or (130,95)" severity error;

		col_base <= "000011110"; row <= "001110000"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (31,112) or (129,78)" severity error;
		assert green = "000000" report "green 000000 (31,112) or (129,78)" severity error;
		assert blue = "011110" report "blue 011110 (31,112) or (129,78)" severity error;

		col_base <= "000011110"; row <= "010000001"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (32,129) or (128,61)" severity error;
		assert green = "000000" report "green 000000 (32,129) or (128,61)" severity error;
		assert blue = "100001" report "blue 100001 (32,129) or (128,61)" severity error;

		col_base <= "000011110"; row <= "010010010"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (33,146) or (127,44)" severity error;
		assert green = "000000" report "green 000000 (33,146) or (127,44)" severity error;
		assert blue = "100100" report "blue 100100 (33,146) or (127,44)" severity error;

		col_base <= "000011110"; row <= "010100011"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (34,163) or (126,27)" severity error;
		assert green = "000000" report "green 000000 (34,163) or (126,27)" severity error;
		assert blue = "100111" report "blue 100111 (34,163) or (126,27)" severity error;

		col_base <= "000011110"; row <= "010110100"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (35,180) or (125,10)" severity error;
		assert green = "000000" report "green 000000 (35,180) or (125,10)" severity error;
		assert blue = "101011" report "blue 101011 (35,180) or (125,10)" severity error;

		col_base <= "000011110"; row <= "111010101"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (37,-43) or (123,233)" severity error;
		assert green = "011111" report "green 011111 (37,-43) or (123,233)" severity error;
		assert blue = "000101" report "blue 000101 (37,-43) or (123,233)" severity error;

		col_base <= "000011110"; row <= "111100110"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (38,-26) or (122,216)" severity error;
		assert green = "011111" report "green 011111 (38,-26) or (122,216)" severity error;
		assert blue = "000100" report "blue 000100 (38,-26) or (122,216)" severity error;

		col_base <= "000011110"; row <= "111110111"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (39,-9) or (121,199)" severity error;
		assert green = "000000" report "green 000000 (39,-9) or (121,199)" severity error;
		assert blue = "000111" report "blue 000111 (39,-9) or (121,199)" severity error;

		col_base <= "000101000"; row <= "000001000"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (40,8) or (120,182)" severity error;
		assert green = "000000" report "green 000000 (40,8) or (120,182)" severity error;
		assert blue = "001010" report "blue 001010 (40,8) or (120,182)" severity error;

		col_base <= "000101000"; row <= "000011001"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (41,25) or (119,165)" severity error;
		assert green = "000000" report "green 000000 (41,25) or (119,165)" severity error;
		assert blue = "001101" report "blue 001101 (41,25) or (119,165)" severity error;

		col_base <= "000101000"; row <= "000101010"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (42,42) or (118,148)" severity error;
		assert green = "000000" report "green 000000 (42,42) or (118,148)" severity error;
		assert blue = "010001" report "blue 010001 (42,42) or (118,148)" severity error;

		col_base <= "000101000"; row <= "000111011"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (43,59) or (117,131)" severity error;
		assert green = "000000" report "green 000000 (43,59) or (117,131)" severity error;
		assert blue = "010100" report "blue 010100 (43,59) or (117,131)" severity error;

		col_base <= "000101000"; row <= "001001100"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (44,76) or (116,114)" severity error;
		assert green = "000000" report "green 000000 (44,76) or (116,114)" severity error;
		assert blue = "010111" report "blue 010111 (44,76) or (116,114)" severity error;

		col_base <= "000101000"; row <= "001011101"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (45,93) or (115,97)" severity error;
		assert green = "000000" report "green 000000 (45,93) or (115,97)" severity error;
		assert blue = "011010" report "blue 011010 (45,93) or (115,97)" severity error;

		col_base <= "000101000"; row <= "001101110"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (46,110) or (114,80)" severity error;
		assert green = "000000" report "green 000000 (46,110) or (114,80)" severity error;
		assert blue = "011110" report "blue 011110 (46,110) or (114,80)" severity error;

		col_base <= "000101000"; row <= "001111111"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (47,127) or (113,63)" severity error;
		assert green = "000000" report "green 000000 (47,127) or (113,63)" severity error;
		assert blue = "100001" report "blue 100001 (47,127) or (113,63)" severity error;

		col_base <= "000101000"; row <= "010010000"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (48,144) or (112,46)" severity error;
		assert green = "000000" report "green 000000 (48,144) or (112,46)" severity error;
		assert blue = "100100" report "blue 100100 (48,144) or (112,46)" severity error;

		col_base <= "000101000"; row <= "010100001"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (49,161) or (111,29)" severity error;
		assert green = "000000" report "green 000000 (49,161) or (111,29)" severity error;
		assert blue = "100111" report "blue 100111 (49,161) or (111,29)" severity error;

		col_base <= "000110010"; row <= "010110010"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (50,178) or (110,12)" severity error;
		assert green = "000000" report "green 000000 (50,178) or (110,12)" severity error;
		assert blue = "101010" report "blue 101010 (50,178) or (110,12)" severity error;

		col_base <= "000110010"; row <= "111010011"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (52,-45) or (108,235)" severity error;
		assert green = "011111" report "green 011111 (52,-45) or (108,235)" severity error;
		assert blue = "000101" report "blue 000101 (52,-45) or (108,235)" severity error;

		col_base <= "000110010"; row <= "111100100"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (53,-28) or (107,218)" severity error;
		assert green = "011111" report "green 011111 (53,-28) or (107,218)" severity error;
		assert blue = "000100" report "blue 000100 (53,-28) or (107,218)" severity error;

		col_base <= "000110010"; row <= "111110101"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (54,-11) or (106,201)" severity error;
		assert green = "000000" report "green 000000 (54,-11) or (106,201)" severity error;
		assert blue = "000111" report "blue 000111 (54,-11) or (106,201)" severity error;

		col_base <= "000110010"; row <= "000000110"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (55,6) or (105,184)" severity error;
		assert green = "000000" report "green 000000 (55,6) or (105,184)" severity error;
		assert blue = "001010" report "blue 001010 (55,6) or (105,184)" severity error;

		col_base <= "000110010"; row <= "000010111"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (56,23) or (104,167)" severity error;
		assert green = "000000" report "green 000000 (56,23) or (104,167)" severity error;
		assert blue = "001101" report "blue 001101 (56,23) or (104,167)" severity error;

		col_base <= "000110010"; row <= "000101000"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (57,40) or (103,150)" severity error;
		assert green = "000000" report "green 000000 (57,40) or (103,150)" severity error;
		assert blue = "010000" report "blue 010000 (57,40) or (103,150)" severity error;

		col_base <= "000110010"; row <= "000111001"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (58,57) or (102,133)" severity error;
		assert green = "000000" report "green 000000 (58,57) or (102,133)" severity error;
		assert blue = "010011" report "blue 010011 (58,57) or (102,133)" severity error;

		col_base <= "000110010"; row <= "001001010"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (59,74) or (101,116)" severity error;
		assert green = "000000" report "green 000000 (59,74) or (101,116)" severity error;
		assert blue = "010111" report "blue 010111 (59,74) or (101,116)" severity error;

		col_base <= "000111100"; row <= "001011011"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (60,91) or (100,99)" severity error;
		assert green = "000000" report "green 000000 (60,91) or (100,99)" severity error;
		assert blue = "011010" report "blue 011010 (60,91) or (100,99)" severity error;

		col_base <= "000111100"; row <= "001101100"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (61,108) or (99,82)" severity error;
		assert green = "000000" report "green 000000 (61,108) or (99,82)" severity error;
		assert blue = "011101" report "blue 011101 (61,108) or (99,82)" severity error;

		col_base <= "000111100"; row <= "001111101"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (62,125) or (98,65)" severity error;
		assert green = "000000" report "green 000000 (62,125) or (98,65)" severity error;
		assert blue = "100000" report "blue 100000 (62,125) or (98,65)" severity error;

		col_base <= "000111100"; row <= "010001110"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (63,142) or (97,48)" severity error;
		assert green = "000000" report "green 000000 (63,142) or (97,48)" severity error;
		assert blue = "100100" report "blue 100100 (63,142) or (97,48)" severity error;

		col_base <= "000111100"; row <= "010011111"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (64,159) or (96,31)" severity error;
		assert green = "000000" report "green 000000 (64,159) or (96,31)" severity error;
		assert blue = "100111" report "blue 100111 (64,159) or (96,31)" severity error;

		col_base <= "000111100"; row <= "010110000"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (65,176) or (95,14)" severity error;
		assert green = "000000" report "green 000000 (65,176) or (95,14)" severity error;
		assert blue = "101010" report "blue 101010 (65,176) or (95,14)" severity error;

		col_base <= "000111100"; row <= "111010001"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (67,-47) or (93,237)" severity error;
		assert green = "011111" report "green 011111 (67,-47) or (93,237)" severity error;
		assert blue = "000110" report "blue 000110 (67,-47) or (93,237)" severity error;

		col_base <= "000111100"; row <= "111100010"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (68,-30) or (92,220)" severity error;
		assert green = "011111" report "green 011111 (68,-30) or (92,220)" severity error;
		assert blue = "000101" report "blue 000101 (68,-30) or (92,220)" severity error;

		col_base <= "000111100"; row <= "111110011"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (69,-13) or (91,203)" severity error;
		assert green = "000000" report "green 000000 (69,-13) or (91,203)" severity error;
		assert blue = "000110" report "blue 000110 (69,-13) or (91,203)" severity error;

		col_base <= "001000110"; row <= "000000100"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (70,4) or (90,186)" severity error;
		assert green = "000000" report "green 000000 (70,4) or (90,186)" severity error;
		assert blue = "001010" report "blue 001010 (70,4) or (90,186)" severity error;

		col_base <= "001000110"; row <= "000010101"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (71,21) or (89,169)" severity error;
		assert green = "000000" report "green 000000 (71,21) or (89,169)" severity error;
		assert blue = "001101" report "blue 001101 (71,21) or (89,169)" severity error;

		col_base <= "001000110"; row <= "000100110"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (72,38) or (88,152)" severity error;
		assert green = "000000" report "green 000000 (72,38) or (88,152)" severity error;
		assert blue = "010000" report "blue 010000 (72,38) or (88,152)" severity error;

		col_base <= "001000110"; row <= "000110111"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (73,55) or (87,135)" severity error;
		assert green = "000000" report "green 000000 (73,55) or (87,135)" severity error;
		assert blue = "010011" report "blue 010011 (73,55) or (87,135)" severity error;

		col_base <= "001000110"; row <= "001001000"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (74,72) or (86,118)" severity error;
		assert green = "000000" report "green 000000 (74,72) or (86,118)" severity error;
		assert blue = "010110" report "blue 010110 (74,72) or (86,118)" severity error;

		col_base <= "001000110"; row <= "001011001"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (75,89) or (85,101)" severity error;
		assert green = "000000" report "green 000000 (75,89) or (85,101)" severity error;
		assert blue = "011001" report "blue 011001 (75,89) or (85,101)" severity error;

		col_base <= "001000110"; row <= "001101010"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (76,106) or (84,84)" severity error;
		assert green = "000000" report "green 000000 (76,106) or (84,84)" severity error;
		assert blue = "011101" report "blue 011101 (76,106) or (84,84)" severity error;

		col_base <= "001000110"; row <= "001111011"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (77,123) or (83,67)" severity error;
		assert green = "000000" report "green 000000 (77,123) or (83,67)" severity error;
		assert blue = "100000" report "blue 100000 (77,123) or (83,67)" severity error;

		col_base <= "001000110"; row <= "010001100"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (78,140) or (82,50)" severity error;
		assert green = "000000" report "green 000000 (78,140) or (82,50)" severity error;
		assert blue = "100011" report "blue 100011 (78,140) or (82,50)" severity error;

		col_base <= "001000110"; row <= "010011101"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (79,157) or (81,33)" severity error;
		assert green = "000000" report "green 000000 (79,157) or (81,33)" severity error;
		assert blue = "100110" report "blue 100110 (79,157) or (81,33)" severity error;

		col_base <= "001010000"; row <= "010101110"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (80,174) or (80,16)" severity error;
		assert green = "000000" report "green 000000 (80,174) or (80,16)" severity error;
		assert blue = "101010" report "blue 101010 (80,174) or (80,16)" severity error;

		col_base <= "001010000"; row <= "111001111"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (82,-49) or (78,239)" severity error;
		assert green = "011111" report "green 011111 (82,-49) or (78,239)" severity error;
		assert blue = "000110" report "blue 000110 (82,-49) or (78,239)" severity error;

		col_base <= "001010000"; row <= "111100000"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (83,-32) or (77,222)" severity error;
		assert green = "011111" report "green 011111 (83,-32) or (77,222)" severity error;
		assert blue = "000101" report "blue 000101 (83,-32) or (77,222)" severity error;

		col_base <= "001010000"; row <= "111110001"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (84,-15) or (76,205)" severity error;
		assert green = "000000" report "green 000000 (84,-15) or (76,205)" severity error;
		assert blue = "000110" report "blue 000110 (84,-15) or (76,205)" severity error;

		col_base <= "001010000"; row <= "000000010"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (85,2) or (75,188)" severity error;
		assert green = "000000" report "green 000000 (85,2) or (75,188)" severity error;
		assert blue = "001001" report "blue 001001 (85,2) or (75,188)" severity error;

		col_base <= "001010000"; row <= "000010011"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (86,19) or (74,171)" severity error;
		assert green = "000000" report "green 000000 (86,19) or (74,171)" severity error;
		assert blue = "001100" report "blue 001100 (86,19) or (74,171)" severity error;

		col_base <= "001010000"; row <= "000100100"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (87,36) or (73,154)" severity error;
		assert green = "000000" report "green 000000 (87,36) or (73,154)" severity error;
		assert blue = "010000" report "blue 010000 (87,36) or (73,154)" severity error;

		col_base <= "001010000"; row <= "000110101"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (88,53) or (72,137)" severity error;
		assert green = "000000" report "green 000000 (88,53) or (72,137)" severity error;
		assert blue = "010011" report "blue 010011 (88,53) or (72,137)" severity error;

		col_base <= "001010000"; row <= "001000110"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (89,70) or (71,120)" severity error;
		assert green = "000000" report "green 000000 (89,70) or (71,120)" severity error;
		assert blue = "010110" report "blue 010110 (89,70) or (71,120)" severity error;

		col_base <= "001011010"; row <= "001010111"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (90,87) or (70,103)" severity error;
		assert green = "000000" report "green 000000 (90,87) or (70,103)" severity error;
		assert blue = "011001" report "blue 011001 (90,87) or (70,103)" severity error;

		col_base <= "001011010"; row <= "001101000"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (91,104) or (69,86)" severity error;
		assert green = "000000" report "green 000000 (91,104) or (69,86)" severity error;
		assert blue = "011100" report "blue 011100 (91,104) or (69,86)" severity error;

		col_base <= "001011010"; row <= "001111001"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (92,121) or (68,69)" severity error;
		assert green = "000000" report "green 000000 (92,121) or (68,69)" severity error;
		assert blue = "011111" report "blue 011111 (92,121) or (68,69)" severity error;

		col_base <= "001011010"; row <= "010001010"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (93,138) or (67,52)" severity error;
		assert green = "000000" report "green 000000 (93,138) or (67,52)" severity error;
		assert blue = "100011" report "blue 100011 (93,138) or (67,52)" severity error;

		col_base <= "001011010"; row <= "010011011"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (94,155) or (66,35)" severity error;
		assert green = "000000" report "green 000000 (94,155) or (66,35)" severity error;
		assert blue = "100110" report "blue 100110 (94,155) or (66,35)" severity error;

		col_base <= "001011010"; row <= "010101100"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (95,172) or (65,18)" severity error;
		assert green = "000000" report "green 000000 (95,172) or (65,18)" severity error;
		assert blue = "101001" report "blue 101001 (95,172) or (65,18)" severity error;

		col_base <= "001011010"; row <= "010111101"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (96,189) or (64,1)" severity error;
		assert green = "000000" report "green 000000 (96,189) or (64,1)" severity error;
		assert blue = "101100" report "blue 101100 (96,189) or (64,1)" severity error;

		col_base <= "001011010"; row <= "111011110"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "011111" report "red 011111 (98,-34) or (62,224)" severity error;
		assert green = "011111" report "green 011111 (98,-34) or (62,224)" severity error;
		assert blue = "000101" report "blue 000101 (98,-34) or (62,224)" severity error;

		col_base <= "001011010"; row <= "111101111"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (99,-17) or (61,207)" severity error;
		assert green = "000000" report "green 000000 (99,-17) or (61,207)" severity error;
		assert blue = "000110" report "blue 000110 (99,-17) or (61,207)" severity error;

		col_base <= "001100100"; row <= "000000000"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (100,0) or (60,190)" severity error;
		assert green = "000000" report "green 000000 (100,0) or (60,190)" severity error;
		assert blue = "001001" report "blue 001001 (100,0) or (60,190)" severity error;

		col_base <= "001100100"; row <= "000010001"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (101,17) or (59,173)" severity error;
		assert green = "000000" report "green 000000 (101,17) or (59,173)" severity error;
		assert blue = "001100" report "blue 001100 (101,17) or (59,173)" severity error;

		col_base <= "001100100"; row <= "000100010"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (102,34) or (58,156)" severity error;
		assert green = "000000" report "green 000000 (102,34) or (58,156)" severity error;
		assert blue = "001111" report "blue 001111 (102,34) or (58,156)" severity error;

		col_base <= "001100100"; row <= "000110011"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (103,51) or (57,139)" severity error;
		assert green = "000000" report "green 000000 (103,51) or (57,139)" severity error;
		assert blue = "010010" report "blue 010010 (103,51) or (57,139)" severity error;

		col_base <= "001100100"; row <= "001000100"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (104,68) or (56,122)" severity error;
		assert green = "000000" report "green 000000 (104,68) or (56,122)" severity error;
		assert blue = "010110" report "blue 010110 (104,68) or (56,122)" severity error;

		col_base <= "001100100"; row <= "001010101"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (105,85) or (55,105)" severity error;
		assert green = "000000" report "green 000000 (105,85) or (55,105)" severity error;
		assert blue = "011001" report "blue 011001 (105,85) or (55,105)" severity error;

		col_base <= "001100100"; row <= "001100110"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (106,102) or (54,88)" severity error;
		assert green = "000000" report "green 000000 (106,102) or (54,88)" severity error;
		assert blue = "011100" report "blue 011100 (106,102) or (54,88)" severity error;

		col_base <= "001100100"; row <= "001110111"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (107,119) or (53,71)" severity error;
		assert green = "000000" report "green 000000 (107,119) or (53,71)" severity error;
		assert blue = "011111" report "blue 011111 (107,119) or (53,71)" severity error;

		col_base <= "001100100"; row <= "010001000"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (108,136) or (52,54)" severity error;
		assert green = "000000" report "green 000000 (108,136) or (52,54)" severity error;
		assert blue = "100010" report "blue 100010 (108,136) or (52,54)" severity error;

		col_base <= "001100100"; row <= "010011001"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (109,153) or (51,37)" severity error;
		assert green = "000000" report "green 000000 (109,153) or (51,37)" severity error;
		assert blue = "100101" report "blue 100101 (109,153) or (51,37)" severity error;

		col_base <= "001101110"; row <= "010101010"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (110,170) or (50,20)" severity error;
		assert green = "000000" report "green 000000 (110,170) or (50,20)" severity error;
		assert blue = "101001" report "blue 101001 (110,170) or (50,20)" severity error;

		col_base <= "001101110"; row <= "010111011"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (111,187) or (49,3)" severity error;
		assert green = "000000" report "green 000000 (111,187) or (49,3)" severity error;
		assert blue = "101100" report "blue 101100 (111,187) or (49,3)" severity error;

		col_base <= "001101110"; row <= "111011100"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (113,-36) or (47,226)" severity error;
		assert green = "000000" report "green 000000 (113,-36) or (47,226)" severity error;
		assert blue = "000010" report "blue 000010 (113,-36) or (47,226)" severity error;

		col_base <= "001101110"; row <= "111101101"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (114,-19) or (46,209)" severity error;
		assert green = "000000" report "green 000000 (114,-19) or (46,209)" severity error;
		assert blue = "000101" report "blue 000101 (114,-19) or (46,209)" severity error;

		col_base <= "001101110"; row <= "111111110"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (115,-2) or (45,192)" severity error;
		assert green = "000000" report "green 000000 (115,-2) or (45,192)" severity error;
		assert blue = "001001" report "blue 001001 (115,-2) or (45,192)" severity error;

		col_base <= "001101110"; row <= "000001111"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (116,15) or (44,175)" severity error;
		assert green = "000000" report "green 000000 (116,15) or (44,175)" severity error;
		assert blue = "001100" report "blue 001100 (116,15) or (44,175)" severity error;

		col_base <= "001101110"; row <= "000100000"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (117,32) or (43,158)" severity error;
		assert green = "000000" report "green 000000 (117,32) or (43,158)" severity error;
		assert blue = "001111" report "blue 001111 (117,32) or (43,158)" severity error;

		col_base <= "001101110"; row <= "000110001"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (118,49) or (42,141)" severity error;
		assert green = "000000" report "green 000000 (118,49) or (42,141)" severity error;
		assert blue = "010010" report "blue 010010 (118,49) or (42,141)" severity error;

		col_base <= "001101110"; row <= "001000010"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (119,66) or (41,124)" severity error;
		assert green = "000000" report "green 000000 (119,66) or (41,124)" severity error;
		assert blue = "010101" report "blue 010101 (119,66) or (41,124)" severity error;

		col_base <= "001111000"; row <= "001010011"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (120,83) or (40,107)" severity error;
		assert green = "000000" report "green 000000 (120,83) or (40,107)" severity error;
		assert blue = "011000" report "blue 011000 (120,83) or (40,107)" severity error;

		col_base <= "001111000"; row <= "001100100"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (121,100) or (39,90)" severity error;
		assert green = "000000" report "green 000000 (121,100) or (39,90)" severity error;
		assert blue = "011100" report "blue 011100 (121,100) or (39,90)" severity error;

		col_base <= "001111000"; row <= "001110101"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (122,117) or (38,73)" severity error;
		assert green = "000000" report "green 000000 (122,117) or (38,73)" severity error;
		assert blue = "011111" report "blue 011111 (122,117) or (38,73)" severity error;

		col_base <= "001111000"; row <= "010000110"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (123,134) or (37,56)" severity error;
		assert green = "000000" report "green 000000 (123,134) or (37,56)" severity error;
		assert blue = "100010" report "blue 100010 (123,134) or (37,56)" severity error;

		col_base <= "001111000"; row <= "010010111"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (124,151) or (36,39)" severity error;
		assert green = "000000" report "green 000000 (124,151) or (36,39)" severity error;
		assert blue = "100101" report "blue 100101 (124,151) or (36,39)" severity error;

		col_base <= "001111000"; row <= "010101000"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (125,168) or (35,22)" severity error;
		assert green = "000000" report "green 000000 (125,168) or (35,22)" severity error;
		assert blue = "101000" report "blue 101000 (125,168) or (35,22)" severity error;

		col_base <= "001111000"; row <= "010111001"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (126,185) or (34,5)" severity error;
		assert green = "000000" report "green 000000 (126,185) or (34,5)" severity error;
		assert blue = "101011" report "blue 101011 (126,185) or (34,5)" severity error;

		col_base <= "001111000"; row <= "111011010"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (128,-38) or (32,228)" severity error;
		assert green = "000000" report "green 000000 (128,-38) or (32,228)" severity error;
		assert blue = "000010" report "blue 000010 (128,-38) or (32,228)" severity error;

		col_base <= "001111000"; row <= "111101011"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (129,-21) or (31,211)" severity error;
		assert green = "000000" report "green 000000 (129,-21) or (31,211)" severity error;
		assert blue = "000101" report "blue 000101 (129,-21) or (31,211)" severity error;

		col_base <= "010000010"; row <= "111111100"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (130,-4) or (30,194)" severity error;
		assert green = "000000" report "green 000000 (130,-4) or (30,194)" severity error;
		assert blue = "001000" report "blue 001000 (130,-4) or (30,194)" severity error;

		col_base <= "010000010"; row <= "000001101"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (131,13) or (29,177)" severity error;
		assert green = "000000" report "green 000000 (131,13) or (29,177)" severity error;
		assert blue = "001011" report "blue 001011 (131,13) or (29,177)" severity error;

		col_base <= "010000010"; row <= "000011110"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (132,30) or (28,160)" severity error;
		assert green = "000000" report "green 000000 (132,30) or (28,160)" severity error;
		assert blue = "001111" report "blue 001111 (132,30) or (28,160)" severity error;

		col_base <= "010000010"; row <= "000101111"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (133,47) or (27,143)" severity error;
		assert green = "000000" report "green 000000 (133,47) or (27,143)" severity error;
		assert blue = "010010" report "blue 010010 (133,47) or (27,143)" severity error;

		col_base <= "010000010"; row <= "001000000"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (134,64) or (26,126)" severity error;
		assert green = "000000" report "green 000000 (134,64) or (26,126)" severity error;
		assert blue = "010101" report "blue 010101 (134,64) or (26,126)" severity error;

		col_base <= "010000010"; row <= "001010001"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (135,81) or (25,109)" severity error;
		assert green = "000000" report "green 000000 (135,81) or (25,109)" severity error;
		assert blue = "011000" report "blue 011000 (135,81) or (25,109)" severity error;

		col_base <= "010000010"; row <= "001100010"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (136,98) or (24,92)" severity error;
		assert green = "000000" report "green 000000 (136,98) or (24,92)" severity error;
		assert blue = "011011" report "blue 011011 (136,98) or (24,92)" severity error;

		col_base <= "010000010"; row <= "001110011"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (137,115) or (23,75)" severity error;
		assert green = "000000" report "green 000000 (137,115) or (23,75)" severity error;
		assert blue = "011110" report "blue 011110 (137,115) or (23,75)" severity error;

		col_base <= "010000010"; row <= "010000100"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (138,132) or (22,58)" severity error;
		assert green = "000000" report "green 000000 (138,132) or (22,58)" severity error;
		assert blue = "100010" report "blue 100010 (138,132) or (22,58)" severity error;

		col_base <= "010000010"; row <= "010010101"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (139,149) or (21,41)" severity error;
		assert green = "000000" report "green 000000 (139,149) or (21,41)" severity error;
		assert blue = "100101" report "blue 100101 (139,149) or (21,41)" severity error;

		col_base <= "010001100"; row <= "010100110"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (140,166) or (20,24)" severity error;
		assert green = "000000" report "green 000000 (140,166) or (20,24)" severity error;
		assert blue = "101000" report "blue 101000 (140,166) or (20,24)" severity error;

		col_base <= "010001100"; row <= "010110111"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (141,183) or (19,7)" severity error;
		assert green = "000000" report "green 000000 (141,183) or (19,7)" severity error;
		assert blue = "101011" report "blue 101011 (141,183) or (19,7)" severity error;

		col_base <= "010001100"; row <= "111011000"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (143,-40) or (17,230)" severity error;
		assert green = "000000" report "green 000000 (143,-40) or (17,230)" severity error;
		assert blue = "000001" report "blue 000001 (143,-40) or (17,230)" severity error;

		col_base <= "010001100"; row <= "111101001"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (144,-23) or (16,213)" severity error;
		assert green = "000000" report "green 000000 (144,-23) or (16,213)" severity error;
		assert blue = "000100" report "blue 000100 (144,-23) or (16,213)" severity error;

		col_base <= "010001100"; row <= "111111010"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (145,-6) or (15,196)" severity error;
		assert green = "000000" report "green 000000 (145,-6) or (15,196)" severity error;
		assert blue = "001000" report "blue 001000 (145,-6) or (15,196)" severity error;

		col_base <= "010001100"; row <= "000001011"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (146,11) or (14,179)" severity error;
		assert green = "000000" report "green 000000 (146,11) or (14,179)" severity error;
		assert blue = "001011" report "blue 001011 (146,11) or (14,179)" severity error;

		col_base <= "010001100"; row <= "000011100"; col_offset <= "0111";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (147,28) or (13,162)" severity error;
		assert green = "000000" report "green 000000 (147,28) or (13,162)" severity error;
		assert blue = "001110" report "blue 001110 (147,28) or (13,162)" severity error;

		col_base <= "010001100"; row <= "000101101"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (148,45) or (12,145)" severity error;
		assert green = "000000" report "green 000000 (148,45) or (12,145)" severity error;
		assert blue = "010001" report "blue 010001 (148,45) or (12,145)" severity error;

		col_base <= "010001100"; row <= "000111110"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (149,62) or (11,128)" severity error;
		assert green = "000000" report "green 000000 (149,62) or (11,128)" severity error;
		assert blue = "010101" report "blue 010101 (149,62) or (11,128)" severity error;

		col_base <= "010010110"; row <= "001001111"; col_offset <= "0000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (150,79) or (10,111)" severity error;
		assert green = "000000" report "green 000000 (150,79) or (10,111)" severity error;
		assert blue = "011000" report "blue 011000 (150,79) or (10,111)" severity error;

		col_base <= "010010110"; row <= "001100000"; col_offset <= "0001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (151,96) or (9,94)" severity error;
		assert green = "000000" report "green 000000 (151,96) or (9,94)" severity error;
		assert blue = "011011" report "blue 011011 (151,96) or (9,94)" severity error;

		col_base <= "010010110"; row <= "001110001"; col_offset <= "0010";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (152,113) or (8,77)" severity error;
		assert green = "000000" report "green 000000 (152,113) or (8,77)" severity error;
		assert blue = "011110" report "blue 011110 (152,113) or (8,77)" severity error;

		col_base <= "010010110"; row <= "010000010"; col_offset <= "0011";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (153,130) or (7,60)" severity error;
		assert green = "000000" report "green 000000 (153,130) or (7,60)" severity error;
		assert blue = "100001" report "blue 100001 (153,130) or (7,60)" severity error;

		col_base <= "010010110"; row <= "010010011"; col_offset <= "0100";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (154,147) or (6,43)" severity error;
		assert green = "000000" report "green 000000 (154,147) or (6,43)" severity error;
		assert blue = "100100" report "blue 100100 (154,147) or (6,43)" severity error;

		col_base <= "010010110"; row <= "010100100"; col_offset <= "0101";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (155,164) or (5,26)" severity error;
		assert green = "000000" report "green 000000 (155,164) or (5,26)" severity error;
		assert blue = "101000" report "blue 101000 (155,164) or (5,26)" severity error;

		col_base <= "010010110"; row <= "010110101"; col_offset <= "0110";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (156,181) or (4,9)" severity error;
		assert green = "000000" report "green 000000 (156,181) or (4,9)" severity error;
		assert blue = "101011" report "blue 101011 (156,181) or (4,9)" severity error;

		col_base <= "010010110"; row <= "111010110"; col_offset <= "1000";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (158,-42) or (2,232)" severity error;
		assert green = "000000" report "green 000000 (158,-42) or (2,232)" severity error;
		assert blue = "000001" report "blue 000001 (158,-42) or (2,232)" severity error;

		col_base <= "010010110"; row <= "111100111"; col_offset <= "1001";
		new_pixel <= '1'; wait for 20 ns; new_pixel <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		rec_x0 <= "000000000"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "011001000"; rec_y1 <= "011001000"; rec_z1 <= "000000000";
		color <= "111100"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		rec_x0 <= "000110010"; rec_y0 <= "000000000"; rec_z0 <= "000000000";
		rec_x1 <= "000110010"; rec_y1 <= "001100100"; rec_z1 <= "011111110";
		color <= "110000"; new_rectangle <= '1'; wait for 20 ns;
		new_rectangle <= '0'; wait until done = '1'; wait for 20 ns;
		done_rectangles <= '1'; wait for 20 ns; done_rectangles <= '0';
		assert red = "000000" report "red 000000 (159,-25) or (1,215)" severity error;
		assert green = "000000" report "green 000000 (159,-25) or (1,215)" severity error;
		assert blue = "000100" report "blue 000100 (159,-25) or (1,215)" severity error;
		
		report "Don't worry, it's just the end of the simulation." severity failure;
	end process;
end ray_test;
